import math
# pip install midiutil
from midiutil import MIDIFile
exec(open("tune_data.py").read())
clk_freq = 3500000

# single-use code for turning Tim Follin's Star Tip 2 into MIDI
# written by Ben Supper
# where we're going, we don't need a licence: hack away.
# -----------------------------------------------------------------------------

# runs an attack, decay, or sustain envelope simulation, counting up T-states
# and counting down remaining duration (in loop cycles) as it goes. ends
# after maxIterations.
def runEnvelope(duration, rate, maxIterations, tWhenIdle, tWhenActive):
    cycles = rate
    t = 0
    iterations = maxIterations
    while (duration >= 10) and (iterations > 0):
        cycles -= 1
        if (cycles == 0):
            cycles = rate
            iterations -= 1
            t += tWhenActive
        else:
            t += tWhenIdle
        #
        # E == envelope skip == 10
        duration -= 10
    #
    return [duration, t]
#

# -----------------------------------------------------------------------------

# calculate the total number of T-states that turn over when a new chord is played.
def tStatesInChord(duration, atk, dcy, sustain, note0, note1, note2):
    # 28: chord_loop setup
    # 48: 3 notes * 16 T-states per skipped note
    # 14: test register E
    # 34: skip_envelope
    start_duration = duration
    # -1 is a fiddle factor: otherwise, everything's a bit flat and takes
    # a bit long: I have probably, in spite of my care, miscounted somewhere
    t_idle = (28+48+14+34-1) * duration
    # then account for each note
    # conditional part of JR runs a bit quicker, then we have a 260 T-state routine for outputting a pulse
    t_per_note = 260-5
    t_notes  = t_per_note * (duration // note0)
    t_notes += t_per_note * (duration // note1)
    t_notes += t_per_note * (duration // note2)
    # envelope load done in a loop
    # attack
    [duration, t_atk] = runEnvelope(duration, atk, 14, 34+40, 34+40+26+47) # 1..15 so 14 iterations
    if (duration > 0): t_atk += 44 # max out; entering decay mode
    # decay
    [duration, t_dcy] = runEnvelope(duration, dcy, 15-sustain, 34+40, 34+40+26+48+17)
    # sustain: it's decay less the write-back
    [duration, t_sustain] = runEnvelope(duration, dcy, 2**20, 34+40, 34+40+26+48)
    return [t_idle, t_notes, t_atk, t_dcy, t_sustain]
#

# -----------------------------------------------------------------------------

# inter-interval pauses, dictated entirely by the time it takes for the program
# to maintain its internal state.
def tStatesBetweenNotes(new_envelope):
    # 10: RET
    # 36: post chord_loop keyboard scan
    # 63: post chord_loop pointer advance and termination check
    # 33: beginning of main_loop
    if new_envelope:
        # just a new envelope: a new note sends us around the loop again
        return 33+220 # the note ends too, so we're just adding the part at the
                      # start of main_loop
    else:
        # in-between notes
        return 10+36+63+33+186
    #
# 

# -----------------------------------------------------------------------------

# convert pitch (in Hz) to MIDI note
def ftn(freq):
    note = freq / 440
    note = math.log2(note) # 440Hz -> 0; 880Hz -> 1
    note = note * 12 + 69 # 440Hz -> 69; 880Hz -> 81
    return note
#

# -----------------------------------------------------------------------------

# add a note to a MIDI file
def addNote(midi, pitch, time_beats, dur_beats, includePitchBend):
    bendrange = 2
    irp = int(round(pitch))
    if includePitchBend:
        # 2 semitone range
        frac = pitch-irp
        bend = ((8192*frac)/bendrange) #+8192
        midi.addPitchWheelEvent(0, 0, time_beats-0.01, int(round(bend)))
    #
    midi.addNote(0, 0, irp, time_beats, dur_beats, 100)
    return midi
#

# -----------------------------------------------------------------------------

def addExpression(midi, expression):
    # MIDI CC7 = volume
    CC = 7
    # expression == [time, dur_secs, fPerCycle, atk, dcy, sustain]
    envTick = 10/expression[2]
    time = 2*expression[0]
    endtime = 2*(expression[0]+expression[1])
    # attack
    for i in range(1,15): # ends at 14
        if (time < endtime):
            midi.addControllerEvent(0, 0, time, CC, 6*i - 2)
            time += 2 * envTick * expression[3]
        #
    #
    # decay
    for i in range(15,expression[5]-1,-1):
        if (time < endtime):
            midi.addControllerEvent(0, 0, time, CC, 6*i - 2)
            time += 2 * envTick * expression[4]
        #
    #
    return midi
#

# -----------------------------------------------------------------------------

# write a MIDI file given a list of time, duration, and MIDI note entries
def writeMidi(filename, solo, notedata, expressiondata):
    midi = MIDIFile(1)
    midi.addTempo(1, 0, 120) # 120bpm is the DAW defaultclk_freq = 3500000
    for note in notedata:
        time_secs = note[0] + 0.01
        dur_secs = note[1]
        # as the tempo is 120bpm, multiply time by 2 and let the notes detach very slightly
        # (by about 1/12th of the shortest note duration) as it's good form
        if (solo == 0) or (solo == 1): midi = addNote(midi, note[2], 2*time_secs, 2*dur_secs-0.02, solo != 0)
        if (solo == 0) or (solo == 2): midi = addNote(midi, note[3], 2*time_secs, 2*dur_secs-0.02, solo != 0)
        if (solo == 0) or (solo == 3): midi = addNote(midi, note[4], 2*time_secs, 2*dur_secs-0.02, solo != 0)

    # in expression data, every channel is treated the same
    for expression in expressiondata:
        midi = addExpression(midi, expression)
    #
    
    with open(filename,"wb") as output_file:
        midi.writeFile(output_file)
    #
#

# -----------------------------------------------------------------------------
# script starts here

# CSV file column names
TIME = 0
DURATIONSECS = 1
DURATIONLOOPS = 2
ATK = 3
DCY = 4
SUSTAIN = 5
TIDLE = 6
TNOTES = 7
TATK = 8
TDCY = 9
TSUST = 10
TGAP = 11
TTOTAL = 12
MIDI0 = 13
MIDI1 = 14
MIDI2 = 15
COLS = 16
csv = "time/s,duration less gap/s,duration less gap/loops,atk,dcy,sustain,t_idle,t_notes,t_atk,t_dcy,t_sustain,t_gap,t_total,MIDI note 0,MIDI note 1,MIDI note 2\n"
# chord is used to compile data from the most recent chord,
# so it can be output a line at a time
chord = [0] * COLS
# the notedata list compiles only five columns (time, duration, and three notes)
# but every note is collected into it so that the differently-specified
# MIDI files can all be run off at the end.
notedata = []
# the expressiondata list contains the envelopes and timings required to output
# volume data for each note
expressiondata = []

# -----------------------------------------------------------------------------
# note extraction loop

time = 0
p = 0
while p < len(tune_data):
    t_gap = 0
    if (tune_data[p] == 0xff):
        # new envelope
        dur_loops = tune_data[p+1] + 256*tune_data[p+2]
        chord[DURATIONLOOPS] = dur_loops
        #
        atk = tune_data[p+3]
        if atk == 0: atk = 256
        chord[ATK] = atk
        #
        dcy = tune_data[p+4]
        if dcy == 0: dcy = 256
        chord[DCY] = dcy
        #
        sustain = tune_data[p+5]
        chord[SUSTAIN] = sustain
        #
        if (time > 0): t_gap = tStatesBetweenNotes(True)
        p += 6
    #
    if (p < len(tune_data)) and (tune_data[p] != 0xff):
        pitch0 = tune_data[p+0]
        pitch1 = tune_data[p+1]
        pitch2 = tune_data[p+2]
        [t_idle, t_notes, t_atk, t_dcy, t_sustain] = tStatesInChord(dur_loops, atk, dcy, sustain, pitch0, pitch1, pitch2)
        if (time > 0): t_gap += tStatesBetweenNotes(False)
        t_total = t_idle + t_notes + t_atk + t_dcy + t_sustain + t_gap
        #
        chord[TIDLE] = t_idle
        chord[TNOTES] = t_notes
        chord[TATK] = t_atk
        chord[TDCY] = t_dcy
        chord[TSUST] = t_sustain
        chord[TGAP] = t_gap
        chord[TTOTAL] = t_total
        #
        tPerCycle = (t_total-t_gap) / dur_loops
        fPerCycle = clk_freq / tPerCycle
        note0 = ftn(fPerCycle / pitch0)
        note1 = ftn(fPerCycle / pitch1)
        note2 = ftn(fPerCycle / pitch2)
        #
        chord[MIDI0] = note0
        chord[MIDI1] = note1
        chord[MIDI2] = note2
        #
        dur_secs = (t_total-t_gap) / clk_freq
        chord[TIME] = time
        chord[DURATIONSECS] = dur_secs
        #
        notedata.append([time, dur_secs, note0, note1, note2])
        expressiondata.append([time, dur_secs, fPerCycle, atk, dcy, sustain])
        csv += "%f,%f,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%f,%f,%f\n" % tuple(chord)
        #
        time += t_total / clk_freq # including gap
        #
        p += 3
    #
#

writeMidi("follin123.mid", 0, notedata, expressiondata)
writeMidi("follin1.mid", 1, notedata, expressiondata)
writeMidi("follin2.mid", 2, notedata, expressiondata)
writeMidi("follin3.mid", 3, notedata, expressiondata)
#
open("follin.csv","w").write(csv)
#
